package service

import (
	"context"
	"fmt"
	"regexp"

	"github.com/ttacon/libphonenumber"
	account "gitlab.mashroom.com/mashroom/backend/account/api"
	auth "gitlab.mashroom.com/mashroom/backend/auth/api"
	"gitlab.mashroom.com/mashroom/backend/auth/internal/storage"
	invitation "gitlab.mashroom.com/mashroom/backend/invitation/api"
	messaging "gitlab.mashroom.com/mashroom/backend/messaging/api"
	notification "gitlab.mashroom.com/mashroom/backend/notification/api"
)

// Sensitivity represents a level of required checks for update action.
type Sensitivity int32

// Provided Sensitivity value
const (
	SensitivityVerification Sensitivity = 1 << iota
	SensitivityOtpEmail
	SensitivityOtpPhone
	SensitivityOtpNewPhone
	SensitivityOtpNewEmail
)

func determineSensitivity(acc *account.Account, password string) (res Sensitivity) {
	if password != "" {
		res |= SensitivityVerification
	}

	if acc == nil {
		return res
	}

	if acc.Contact != nil {
		contact := acc.GetContactValue()

		if contact.GetEmailPrimary() != nil {
			res |= SensitivityVerification | SensitivityOtpNewEmail | SensitivityOtpPhone
		}

		if contact.GetPhonePrimary() != nil {
			res |= SensitivityOtpNewPhone
		}
	}

	return res
}

func (srv *serverImpl) validatePassword(password string) (errs []auth.AuthErrors) {
	if len(password) < 8 {
		errs = append(errs, auth.AuthErrors_AUTH_ERROR_PASSWORD_TOO_SHORT)
	}

	if len(password) > 32 {
		errs = append(errs, auth.AuthErrors_AUTH_ERROR_PASSWORD_TOO_LONG)
	}

	return
}

func (srv *serverImpl) validateNewPassword(password, newPassword string) (errs []auth.AuthErrors) {
	if newPassword != "" && password == "" {
		errs = append(errs, auth.AuthErrors_AUTH_ERROR_PASSWORD_TOO_SHORT)
	}

	return
}

func (srv *serverImpl) validateTosAccepted(tosAccepted uint64) (errs []auth.AuthErrors) {
	if tosAccepted == 0 {
		errs = append(errs, auth.AuthErrors_AUTH_ERROR_TOS_REVIEW_REQUIRED)
	}

	return
}

func (srv *serverImpl) validateEmailUnique(
	ctx context.Context,
	email, userID string,
) (errs []auth.AuthErrors, err error) {
	var users []*storage.User

	users, err = srv.Storage.UserGetByEmail(ctx, email)
	if err != nil {
		return
	}

	for _, u := range users {
		if u.Id != userID {
			errs = append(errs, auth.AuthErrors_AUTH_ERROR_EMAIL_TAKEN)
			break
		}
	}

	return
}

func (srv *serverImpl) validateAccount(
	ctx context.Context,
	acc *account.Account,
) (valid bool, verif *account.AccountVerification, err error) {
	resp, err := srv.Connections.Account.Verify(ctx, &account.VerifyRequest{
		Account: acc,
	})
	if err != nil {
		return
	}

	valid, verif = resp.GetValid(), resp.GetVerification()

	return
}

func (srv *serverImpl) validateOTP(
	ctx context.Context,
	userID, key, otp string,
	typ invitation.ContactValidationType,
	status *account.AccountContactStatus,
	req *notification.SendNotificationRequest,
) (errs []auth.AuthErrors, err error) {
	if key == "" {
		return
	}

	cvresp, err := srv.Connections.Invitation.ContactValidationCreate(
		ctx,
		&invitation.ContactValidationCreateRequest{
			Type:   typ,
			UserId: userID,
			Key:    key,
		},
	)
	if err != nil {
		return
	}

	if req != nil {
		req.Embeds = append(
			req.Embeds,
			&notification.EventEmbed{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_TOKEN,
				EmbedValue: cvresp.GetContactValidation().GetOtp(),
			},
		)

		_, err = srv.Connections.Notification.SendNotification(ctx, req)
		if err != nil {
			return nil, err
		}
	}

	if otp == "" {
		switch typ {
		case invitation.ContactValidationType_CONTACT_VALIDATION_PHONE:
			errs = append(errs, auth.AuthErrors_AUTH_ERROR_PHONE_OTP_REQUIRED)
		case invitation.ContactValidationType_CONTACT_VALIDATION_EMAIL:
			errs = append(errs, auth.AuthErrors_AUTH_ERROR_EMAIL_OTP_REQUIRED)
		}

		return
	}

	resp, err := srv.Connections.Invitation.ContactValidationVerify(
		ctx,
		&invitation.ContactValidationVerifyRequest{
			Type:   typ,
			UserId: userID,
			Key:    key,
			Otp:    otp,
		},
	)
	if err != nil {
		return
	}

	if resp.Status == invitation.ContactValidationStatus_CONTACT_VALIDATION_STATUS_OK {
		switch typ {
		case invitation.ContactValidationType_CONTACT_VALIDATION_PHONE:
			status.PhoneVerified = &account.AccountContactStatus_PhoneVerifiedValue{PhoneVerifiedValue: true}
		case invitation.ContactValidationType_CONTACT_VALIDATION_EMAIL:
			status.EmailVerified = &account.AccountContactStatus_EmailVerifiedValue{EmailVerifiedValue: true}
		}

		return
	}

	switch typ {
	case invitation.ContactValidationType_CONTACT_VALIDATION_PHONE:
		errs = append(errs, auth.AuthErrors_AUTH_ERROR_PHONE_OTP_INVALID)
		status.PhoneVerified = &account.AccountContactStatus_PhoneVerifiedValue{PhoneVerifiedValue: false}
	case invitation.ContactValidationType_CONTACT_VALIDATION_EMAIL:
		errs = append(errs, auth.AuthErrors_AUTH_ERROR_EMAIL_OTP_INVALID)
		status.EmailVerified = &account.AccountContactStatus_EmailVerifiedValue{EmailVerifiedValue: false}
	}

	return
}

func (srv *serverImpl) validateUpdateAccount(
	ctx context.Context,
	introspect *storage.SessionIntrospection,
	current, update *account.Account,
	password, newPassword, pin, phoneOTP, emailOTP string,
) (valid bool, verif *account.AccountVerification, autherrs []auth.AuthErrors, err error) {
	autherrs = append(autherrs, srv.validateNewPassword(password, newPassword)...)
	if len(autherrs) > 0 {
		return
	}

	valid, verif, err = srv.validateAccount(ctx, update)
	if err != nil {
		return
	}

	email := update.GetContactValue().GetEmailPrimaryValue()
	phone := update.GetContactValue().GetPhonePrimaryValue()

	if newPassword != "" {
		pin = ""

		autherrs = srv.validatePassword(newPassword)
	}

	if len(autherrs) > 0 {
		return
	}

	srv.checkSensitivity(
		ctx,
		introspect,
		current,
		update,
		password,
		pin,
		phoneOTP,
		emailOTP,
		email,
		phone,
		&autherrs,
		&err,
	)

	return
}

func (srv *serverImpl) checkSensitivity(
	ctx context.Context,
	introspect *storage.SessionIntrospection,
	current, update *account.Account,
	password, pin, phoneOTP, emailOTP, email, phone string,
	autherrs *[]auth.AuthErrors,
	err *error,
) {
	sens := determineSensitivity(update, password)

	if sens&SensitivityVerification != 0 &&
		!isNewPhoneVerification(current.GetContactValue().GetPhonePrimaryValue(), phone) {
		*autherrs = srv.verifyUserIntent(ctx, password, pin, keycloakTokenIdentity(introspect), introspect.Sub)
	}

	if len(*autherrs) > 0 {
		return
	}

	if sens&SensitivityOtpNewEmail != 0 {
		*autherrs, *err = srv.validateEmailUnique(ctx, email, introspect.Sub)
		if *err != nil {
			return
		}
	}

	if len(*autherrs) > 0 {
		return
	}

	if sens&SensitivityOtpNewEmail != 0 {
		*autherrs, *err = srv.validateOTP(
			ctx,
			introspect.Sub,
			email,
			emailOTP,
			invitation.ContactValidationType_CONTACT_VALIDATION_EMAIL,
			update.GetContactValue().GetContactStatusValue(),
			nil,
		)
		if *err != nil {
			return
		}
	}

	if len(*autherrs) > 0 {
		return
	}

	if sens&SensitivityOtpNewPhone != 0 &&
		!isNewPhoneVerification(current.GetContactValue().GetPhonePrimaryValue(), phone) {
		*autherrs, *err = srv.validateOTP(
			ctx,
			introspect.Sub,
			phone,
			phoneOTP,
			invitation.ContactValidationType_CONTACT_VALIDATION_PHONE,
			update.GetContactValue().GetContactStatusValue(),
			nil,
		)
		if *err != nil {
			return
		}
	}

	if len(*autherrs) > 0 {
		return
	}

	if sens&SensitivityOtpEmail != 0 {
		*autherrs, *err = srv.validateOTP(
			ctx,
			introspect.Sub,
			current.GetContactValue().GetEmailPrimaryValue(),
			emailOTP,
			invitation.ContactValidationType_CONTACT_VALIDATION_EMAIL,
			update.GetContactValue().GetContactStatusValue(),
			nil,
		)
		if *err != nil {
			return
		}
	}

	if len(*autherrs) > 0 {
		return
	}

	if sens&SensitivityOtpPhone != 0 {
		*autherrs, *err = srv.validateOTP(
			ctx,
			introspect.Sub,
			current.GetContactValue().GetPhonePrimaryValue(),
			phoneOTP,
			invitation.ContactValidationType_CONTACT_VALIDATION_PHONE,
			update.GetContactValue().GetContactStatusValue(),
			nil,
		)
		if *err != nil {
			return
		}
	}

	if len(*autherrs) > 0 {
		return
	}
}

func (srv *serverImpl) validateCreateAccount(
	ctx context.Context,
	acc *account.Account,
	password, oauthID, oauthType, remote, phoneOTP string,
	adminSite, adminValue bool,
	tosAccepted uint64,
) (valid bool, verif *account.AccountVerification, autherrs []auth.AuthErrors, err error) {
	if oauthID == "" || oauthType == "" || !srv.Config.OAuthEnabled {
		autherrs = append(autherrs, srv.validatePassword(password)...)
	}

	valid, verif, err = srv.validateAccount(ctx, acc)
	if err != nil {
		return
	}

	if !valid {
		return
	}

	var errs []auth.AuthErrors

	errs, err = srv.validateEmailUnique(ctx, acc.GetContactValue().GetEmailPrimaryValue(), "")
	if err != nil {
		return
	}

	autherrs = append(autherrs, errs...)

	if len(autherrs) > 0 {
		return
	}

	if !(adminSite && adminValue) {
		userid := "newuser-" + remote

		errs, err = srv.validateOTP(
			ctx,
			userid,
			acc.GetContactValue().GetPhonePrimaryValue(),
			phoneOTP,
			invitation.ContactValidationType_CONTACT_VALIDATION_PHONE,
			acc.GetContactValue().GetContactStatusValue(),
			nil,
		)
		if err != nil {
			return
		}

		autherrs = append(autherrs, errs...)

		if len(autherrs) > 0 {
			return
		}
	}

	if !adminSite {
		autherrs = append(autherrs, srv.validateTosAccepted(tosAccepted)...)
	}

	return
}

func (srv *serverImpl) validatePasswordReset(
	ctx context.Context,
	acc *account.Account,
	newPassword, emailOTP string,
) (autherrs []auth.AuthErrors, err error) {
	autherrs, err = srv.validateOTP(
		ctx,
		acc.GetIdValue(),
		acc.GetContactValue().GetEmailPrimaryValue(),
		emailOTP,
		invitation.ContactValidationType_CONTACT_VALIDATION_EMAIL,
		acc.GetContactValue().GetContactStatusValue(),
		&notification.SendNotificationRequest{
			Entity: notification.Entity_ENTITY_USER_PASSWORD,
			Action: notification.Action_ACTION_EXPIRE,
			Targets: []*notification.Target{
				{
					User: acc.GetIdValue(),
					Role: notification.Role_ROLE_AUTHOR,
				},
			},
		},
	)
	if err != nil || len(autherrs) > 0 {
		return
	}

	autherrs = srv.validatePassword(newPassword)
	if len(autherrs) > 0 {
		return
	}

	return
}

func isNewPhoneVerification(oldPhone, newPhone string) bool {
	if oldPhone == "" && newPhone != "" {
		return true
	}

	return false
}

var nonDigitReplace = regexp.MustCompile(`[^0-9]`)

func normalizePhoneNumber(inp string) string {
	inp = nonDigitReplace.ReplaceAllString(inp, "")

	parsed, err := libphonenumber.Parse("+"+inp, libphonenumber.UNKNOWN_REGION)
	if err != nil {
		return inp
	}

	return fmt.Sprintf("%d%d", parsed.GetCountryCode(), parsed.GetNationalNumber())
}
